/*
** Module   : XEntityImpl
** Abstract : DOM and SAX common implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130204 Created initial version. Major rework to use w3c backend.
** 002 EVL 20130214 Adding the base inheritance from XCommonImpl class to place the XML related
**                  common features.  
** 003 CA  20130221 Added UNIQUE-ID and NAMESPACE-URI support. 
** 004 EVL 20130222 Adding SUBTYPE attribute and INSERT-BEFORE method support.
** 005 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 006 OM  20130320 Fixed getChild() method, added Override annotations, cosmetics & typo fixes.
*/

package com.goldencode.p2j.xml;

import org.w3c.dom.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of the common features for DOM.
 */
public abstract class XEntityImpl
extends XCommonImpl
implements XEntity
{
   /** The handle associated with this object. */
   private handle hXEntity = null;

   /** Element associates with current XEntity class. */
   private Node xnode = null;

   /**
    * Default constructor.
    */
   public XEntityImpl()
   {
      hXEntity = new handle();
      hXEntity.assign(this);
   }

   /**
    * Special constructor accepting the support object.
    * 
    * @param    value
    *           The object to be associated with the given handle object.
    */
   public XEntityImpl(WrappedResource value)
   {
      this();
      hXEntity.assign(value);
   }

   /**
    * Special constructor accepting a X-Document or X-Node-Reference object.
    *
    * @param    node
    *           The DOM XML node object to be associated with the given
    *           handle object.
    */
   public XEntityImpl(Node node)
   {
      this();
      xnode = node;
      hXEntity.assign(this);
   }

   /**
    * Getting the Element backend object for this XEntity.
    * 
    * @return   The Element associated with object.
    */
   public Node getNode()
   {
      return xnode;
   }
   
   /**
    * Getting the handle of the object.  
    *
    * @return   Internal handle object associated with this node.
    */
   @Override
   public handle getHandle()
   {
      return hXEntity;
   }
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return super.valid() && hXEntity != null && getNode() != null;
   }
   
   /**
    * Reports if this object is unknown.  
    *
    * @return   <code>true</code> if object is unknown.
    */
   @Override
   public boolean unknown()
   {
      return super.unknown() || hXEntity == null || getNode() == null;
   }

//   /**
//    * Perform actual delete of an resource. At the time of this call, it is assumed the resource
//    * is valid for deletion (the handle and the resource are both valid).
//    */
//   @Override
//   public void delete()
//   {
//      //TODO
//   }

   /**
    * Gets the the subtype of the object.
    * 
    * @return   The character value of the subtype attribute.
    */
   @Override
   public character getSubType()
   {
      if (unknown() || !valid())
      {
         return new character();
      }
      final Node node = getNode();
      final character result =
              new character(XmlFactory.convertNodeToAblNode(node.getNodeType()));
      return result;
   }
   
   /**
    * Sets the the subtype of the object.
    * 
    * @param    value
    *           The new value of the subtype object attribute.
    */
   @Override
   public void setSubType(character value)
   {
      readOnlyError("SUBTYPE");
   }

   /**
    * Sets the the subtype of the object.
    * 
    * @param    value
    *           The new value of the subtype object attribute.
    */
   @Override
   public void setSubType(String value)
   {
      setSubType(new character(value));
   }

   /**
    * Inserts a node as a child of this document before another node or last if other node is
    * unknown.
    * 
    * @param    handle1
    *           The handle representing the child to insert.
    * @param    handle2
    *           The handle representing the XML node to be inserted before.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical insertBefore(handle handle1, handle handle2)
   {
      if (unknown() || !valid())
      {
         return new logical(false);
      }
      if (handle2 == null || handle2.isUnknown())
      {
         return appendChild(handle1);
      }
      final Node node2 = ((XEntityImpl) handle2.get()).getNode();
      final Node node1 = ((XEntityImpl) handle1.get()).getNode();
      final Node result = getNode().insertBefore(node1, node2);

      return new logical(result != null);
   }

   /**
    * Gets the the number of children objects for the current one.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getNumChildren()
   {
      if (unknown() || !valid())
      {
         return new integer();
      }
      return new integer(getNode().getChildNodes().getLength());
   }

   /**
    * Removing the specified child from the node.
    *
    * @param newChildHandle The handle class instance that represents the node to insert in the
    *                       tree.
    * @param oldChildHandle The handle class instance that represents the node to remove from the
    *                       tree.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical replaceChild(handle newChildHandle, handle oldChildHandle)
   {
      if (unknown() || !valid())
      {
         return new logical(false);
      }
      if (!newChildHandle._isValid() || !oldChildHandle._isValid())
      {
         return new logical(false);
      }
      final Node newChild = ((XEntityImpl) newChildHandle.get()).getNode();
      final Node oldChild = ((XEntityImpl) oldChildHandle.get()).getNode();
      try
      {
         getNode().replaceChild(newChild, oldChild);
      }
      catch (DOMException e)
      {
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Gets the number of child for the current one. Only for XNodRef
    *
    * @return   The integer number of the child, 1-based index of the child node.
    */
   @Override
   public integer getChildNum()
   {
      if (unknown() || !valid())
      {
         return new integer();
      }
      final Node node = getNode();
      final Node parent = node.getParentNode();
      if (parent == null)
      {
         return new integer();
      }

      final NodeList childNodes = parent.getChildNodes();
      for (int i = 0 ; i < childNodes.getLength(); i++)
      {
         if (node.equals(childNodes.item(i)))
         {
            return new integer(i + 1);
         }
      }

      return new integer();
   }

   /**
    * Appending the new child to the current object.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be appended.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical appendChild(handle hChild)
   {
      if (unknown() || !valid() || !hChild._isValid())
      {
         return new logical(false);
      }

      final Node node = getNode();
      final Node childNode = ((XEntityImpl) hChild.get()).getNode();
      try
      {
          node.appendChild(childNode);
      } catch (DOMException e)
      {
         if (e.code == DOMException.WRONG_DOCUMENT_ERR)
         {
            return new logical(false);
         }
         else
         {
            throw e;
         }
      }
      return new logical(true);
   }

   /**
    * Getting the handle object of the child node with specified index.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to put
    *           the results.
    * @param    index
    *           1-based index of the child node to be retrieved.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical getChild(handle hChild, int index)
   {
      return getChild(hChild, new integer(index));
   }
   
   /**
    * Getting the handle object of the child node with specified index.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to put the results.
    * @param    index
    *           1-based index of the child node to be retrieved.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical getChild(handle hChild, integer index)
   {
      if (unknown() || !valid() || index.isUnknown())
      {
         return new logical(false);
      }

      final int ndx = index.intValue();

      final NodeList nlst = getNode().getChildNodes();
      
      if (ndx < 1 || ndx > nlst.getLength())
      {
         return new logical(false);
      }
      
      final Node node = nlst.item(ndx - 1);
      hChild.assign(new XNodeRefImpl(node));

      return new logical(hChild.isValid());
   }
   
   /**
    * Removing the specified child from the node.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be removed.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical removeChild(handle hChild)
   {
      if (unknown() || !valid() || !hChild._isValid())
      {
         return new logical(false);
      }

      final Node elChild = ((XEntityImpl) hChild.get()).getNode();
      final boolean result = getNode().removeChild(elChild) != null;
      
      return new logical(result);
   }

   /**
    * Clean up the current node.
    */
   public void clear()
   {
      hXEntity = null;
      xnode.getOwnerDocument().removeChild(xnode);
      xnode = null;
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself. For X-noderef it's only unique for current xml Document.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      if (unknown() || !valid())
      {
         return new integer();
      }
      return new integer(getNode().hashCode());
   }
   
   /**
    * Returns the value of the the NAMESPACE-URI attribute.
    *
    * @return    See above.
    */
   @Override
   public character getNamespaceURI()
   {
      final String namespaceUri = getNode().getNamespaceURI();

      if (namespaceUri != null)
      {
         return new character(namespaceUri);
      }
      else
      {
         return new character();
      }
   }

   /**
    * Setter for NAMESPACE-URI attribute. Calls {@link #readOnlyError}, as this attribute is r/o
    * for this resource.
    *
    * @param    uri
    *           Not used.
    */
   @Override
   public void setNamespaceURI(String uri)
   {
      readOnlyError("NAMESPACE-URI");
   }

   /**
    * Setter for NAMESPACE-URI attribute. Calls {@link #readOnlyError}, as this attribute is r/o
    * for this resource.
    *
    * @param    uri
    *           Not used.
    */
   @Override
   public void setNamespaceURI(character uri)
   {
      readOnlyError("NAMESPACE-URI");
   }


   /**
    * Removes the specified attribute.
    *
    * @param    name
    *           The name of the attribute to remove.
    * @param    uri
    *           The URI of the attribute to remove.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical removeAttribute(character name, character uri)
   {
      if (unknown() || !valid())
      {
         return new logical(false);
      }
      final boolean result;
      final Element node = (Element)getNode();
      final String nameStr = name.toStringMessage();
      if (uri.isUnknown())
      {
         result = node.hasAttribute(nameStr);
         node.removeAttribute(nameStr);
      }
      else
      {
         result = node.hasAttributeNS(uri.toStringMessage(), nameStr);
         node.removeAttributeNS(uri.toStringMessage(), nameStr);
      }
      return new logical(result);
   }

   /**
    * Removes the specified attribute. This method is not applicable for Dom.
    *
    * @param    index
    *           The 1-based index of the attribute to remove.
    * @param    uri
    *           The URI of the attribute to remove.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   public logical removeAttribute(NumberType index, character uri)
   {
      return new logical(false); //only for SAX
   }

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @Override
   public void setNamespacePrefix(String prefix)
   {
      setNamespacePrefix(new character(prefix));
   }

   /**
    * Returns the value of the the NAMESPACE-PREFIX attribute. Only for ELEMENT or ATTRIBUTE,
    * for other attributes return unknown value.
    *
    * @return    See above.
    */
   @Override
   public character getNamespacePrefix()
   {
      if (unknown() || !valid())
      {
         return new character();
      }

      final Node currentNode = getNode();
      final String prefix;
      if (currentNode.getNodeType() == Node.ELEMENT_NODE)
      {
         final Element elementNode = (Element) currentNode;
         prefix = elementNode.getPrefix();
      }
      else if (currentNode.getNodeType() == Node.ATTRIBUTE_NODE)
      {
         final Attr attributeNode = (Attr) currentNode;
         prefix = attributeNode.getPrefix();
      }
      else
      {
         prefix = null;
      }

      return new character(prefix);
   }

   /**
    * Get the <code>name</code> attribute of handle.
    *
    * @return   See above.
    */
   public character getName()
   {
      if (unknown() || !valid())
      {
         return new character();
      }

      final Node node = getNode();

      return new character(node.getNodeName());
   }
}
