/*
** Module   : XmlFactory
** Abstract : DOM and SAX XML instance creation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130204 Created initial version. Major rework to use w3c backend.
** 002 EVL 20130208 Added SAX XML Support.
** 003 CA  20130308 The CREATE APIs need to receive a handle parameter, to properly initiate it,
**                  as returning the new handle instance and changing the handle reference in the
**                  caller is not the right approach.
** 004 CA  20130309 Added missing APIs related to WIDGET-POOL clause.
** 005 CA  20130311 Added more missing APIs related to WIDGET-POOL clause.
*/

package com.goldencode.p2j.xml;

import org.w3c.dom.*;
import com.goldencode.p2j.util.*;

import java.util.HashMap;
import java.util.Map;

import static com.goldencode.p2j.xml.XmlTokenTypes.*;

/**
 * The set of methods to work with DOM XML and SAX support implementation in P2J.
 */
public class XmlFactory
{
   /**
    * Constructs a new instance of the Progress X-Document object and
    * retrieves the handle variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    */
   public static void createXDocument(handle h)
   {
      h.assign(new XDocumentImpl());
   }

   /**
    * Constructs a new instance of the Progress X-Document object and
    * retrieves the handle variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    poolName
    *           The name of the pool to which the resource will be added
    */
   public static void createXDocument(handle h, String poolName)
   {
      createXDocument(h, new character(poolName));
   }

   /**
    * Constructs a new instance of the Progress X-Document object and
    * retrieves the handle variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    poolName
    *           The name of the pool to which the resource will be added
    */
   public static void createXDocument(handle h, character poolName)
   {
      h.assign(new XDocumentImpl());
      // TODO: add it to the pool
   }

   /**
    * Constructs a new instance of the Progress X-Node-Reference object and
    * retrieves the handle variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    */
   public static void createXNodeReference(handle h)
   {
      h.assign(new XNodeRefImpl());
   }

   /**
    * Constructs a new instance of the Progress X-Node-Reference object and
    * retrieves the handle variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    poolName
    *           The name of the pool to which the resource will be added
    */
   public static void createXNodeReference(handle h, String poolName)
   {
      createXNodeReference(h, new character(poolName));
   }

   /**
    * Constructs a new instance of the Progress X-Node-Reference object and
    * retrieves the handle variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    poolName
    *           The name of the pool to which the resource will be added
    */
   public static void createXNodeReference(handle h, character poolName)
   {
      h.assign(new XNodeRefImpl());
      // TODO: add it to the pool
   }

   /**
    * Constructs a new instance of the Progress X-Node-Reference object and
    * retrieves the handle variable associated with it.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    doc
    *           The X-Document object to create X-Node under.
    * @param    name
    *           The name of the node to create.
    * @param    type
    *           The type of the node to create.
    */
   public static void createXNodeReference(handle h,
                                           Document doc,
                                           character name,
                                           character type)
   {
      String nameStr = null;
      
      if (!name.isUnknown())
      {
         nameStr = name.toStringMessage();
      }

      int nodeId = convertAblNodeToNodeId(type.toStringMessage());
      final Node node;

      switch (nodeId)
      {
         case ATTRIBUTE_NODE:
            node = doc.createAttribute(nameStr);
            break;
         case CDATA_SECTION_NODE:
            node = doc.createCDATASection("");
            break;
         case COMMENT_NODE:
            node = doc.createComment("");
            break;
         case DOCUMENT_FRAGMENT_NODE:
            node = doc.createDocumentFragment();
            break;
         case ELEMENT_NODE:
            node = doc.createElement(nameStr);
            break;
         case ENTITY_REFERENCE_NODE:
            node = doc.createEntityReference(nameStr);
            break;
         case PROCESSING_INSTRUCTION_NODE:
            node = doc.createProcessingInstruction(nameStr, "");
            break;
         case TEXT_NODE:
            node = doc.createTextNode("");
            break;
         default:
            node = null;
      }

      h.assign(new XNodeRefImpl(node));
   }

   /**
    * Constructs a new instance of the Progress X-Node-Reference object and
    * retrieves the handle variable associated with it. The approach is
    * slightly different. Each node will have additional paranmeter - Name
    * Space and node will be addressed by namespace and name pair instead
    * of the simple name.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    doc
    *           The X-Document object to create X-Node under.
    * @param    nsUri
    *           Name space associated with the current set of names.
    * @param    name
    *           The name of the node to create.
    * @param    type
    *           The type of the node to create.
    */
   public static void createXNodeReference(handle    h,
                                           Document  doc,
                                           character nsUri,
                                           character name,
                                           character type)
   {
      String nameImpl = null;
      String nsUriImpl = null;
      
      if (!nsUri.isUnknown())
      {
         nsUriImpl = nsUri.toStringMessage();
      }
      if (!name.isUnknown())
      {
         nameImpl = name.toStringMessage();
      }
      String typeStr = type.isUnknown() ? "ELEMENT" : "ATTRIBUTE";
      final Node node;
      if ("ELEMENT".equals(typeStr))
      {
         node = doc.createElementNS(nsUriImpl, nameImpl);
      }
      else if ("ATTRIBUTE".equals(typeStr))
      {
         node = doc.createAttributeNS(nsUriImpl, nameImpl);
      }
      else
      {
         node = null;
      }

      h.assign(new XNodeRefImpl(node));
   }

   /**
    * Constructs a new instance of the Progress Sax-Attributes object and retrieves the handle
    * variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    */
   public static void createSaxAttributes(handle h)
   {
      createSaxAttributes(h, new character());
   }

   /**
    * Constructs a new instance of the Progress Sax-Attributes object and retrieves the handle
    * variable associated with it.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    widgetPool
    *           The widget pool in which the object handle is created.
    */
   public static void createSaxAttributes(handle h, String widgetPool)
   {
      createSaxAttributes(h, new character(widgetPool));
   }

   /**
    * Constructs a new instance of the Progress Sax-Attributes object and retrieves the handle
    * variable associated with it.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    widgetPool
    *           The widget pool in which the object handle is created.
    */
   public static void createSaxAttributes(handle h, character widgetPool)
   {
      // TODO: implement this
   }
   
   /**
    * Constructs a new instance of the Progress Sax-Reader object and retrieves the handle
    * variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    */
   public static void createSaxReader(handle h)
   {
      createSaxReader(h, new character());
   }

   /**
    * Constructs a new instance of the Progress Sax-Reader object and retrieves the handle
    * variable associated with it.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    widgetPool
    *           The widget pool in which the object handle is created.
    */
   public static void createSaxReader(handle h, String widgetPool)
   {
      createSaxReader(h, new character(widgetPool));
   }

   /**
    * Constructs a new instance of the Progress Sax-Reader object and retrieves the handle
    * variable associated with it.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    widgetPool
    *           The widget pool in which the object handle is created.
    */
   public static void createSaxReader(handle h, character widgetPool)
   {
      // TODO: implement this
   }
   
   /**
    * Constructs a new instance of the Progress Sax-Writer object and retrieves the handle
    * variable associated with it.
    * 
    * @param    h
    *           The handle where the new resource will be set as the referent.
    */
   public static void createSaxWriter(handle h)
   {
      createSaxWriter(h, new character());
   }

   /**
    * Constructs a new instance of the Progress Sax-Writer object and retrieves the handle
    * variable associated with it.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    widgetPool
    *           The widget pool in which the object handle is created.
    */
   public static void createSaxWriter(handle h, String widgetPool)
   {
      createSaxWriter(h, new character(widgetPool));
   }

   /**
    * Constructs a new instance of the Progress Sax-Writer object and retrieves the handle
    * variable associated with it.
    *
    * @param    h
    *           The handle where the new resource will be set as the referent.
    * @param    widgetPool
    *           The widget pool in which the object handle is created.
    */
   public static void createSaxWriter(handle h, character widgetPool)
   {
      // TODO: implement this
   }

   private static Map<String, Integer> nodeNameToNodeId = new HashMap<String, Integer>();
   private static Map<Integer, String> nodeIdToNodeName = new HashMap<Integer, String>();

   static
   {
      nodeNameToNodeId.put("ATTRIBUTE", ATTRIBUTE_NODE);
      nodeNameToNodeId.put("CDATA-SECTION", CDATA_SECTION_NODE);
      nodeNameToNodeId.put("COMMENT", COMMENT_NODE);
      nodeNameToNodeId.put("DOCUMENT", DOCUMENT_NODE);
      nodeNameToNodeId.put("DOCUMENT-FRAGMENT", DOCUMENT_FRAGMENT_NODE);
      nodeNameToNodeId.put("ELEMENT", ELEMENT_NODE);
      nodeNameToNodeId.put("ENTITY-REFERENCE", ENTITY_REFERENCE_NODE);
      nodeNameToNodeId.put("PROCESSING-INSTRUCTION", PROCESSING_INSTRUCTION_NODE);
      nodeNameToNodeId.put("TEXT", TEXT_NODE);
      for (Map.Entry<String, Integer> entry : nodeNameToNodeId.entrySet())
      {
         nodeIdToNodeName.put(entry.getValue(), entry.getKey());
      }
   }

   //TODO use code from XmlAst class ?
   public static character convertNodeToAblNode(int nodeId)
   {
      return new character(nodeIdToNodeName.get(nodeId));
   }

   public static int convertAblNodeToNodeId(String nodeId)
   {
      final Integer result = nodeNameToNodeId.get(nodeId);
      return result == null ? -1 : result;
   }
}
